function [model status] = bilinsvmtrain(y, X, parms)
%
% Optimize bilinear classification model
%  [model status] = bilinsvmtrain(y, X, params)
%
% min 0.5(|Wh|^2+|Ww|^2 + Wh'Jh Wh + Ww'Jw Ww) + C 1'xi
% s.t. y_i (<x_i, W>_H - b) >= 1 - xi_i
%      xi_i >=0 
%      W = Wh*Ww'
%
% Input:
%  y   - Label vector {-1,+1}  [n x 1]
%  X   - Feature matrices      [h x w x n]
%        NOTE: samples are given in the form of "matrix (h x w)".
%  params - Parameters [struct]
%            .C  - Balancing parameter   [scalar] (default 1)
%            .Jh - Constraint matrix for row weights    [h x h] (default [])
%            .Jw - Constraint matrix for column weights [w x w] (default [])
%		            These matrixes Jh and Jw are used such as 
%		            for smoothing regularizatoin (see sec 3.1 in the paper)
%	                produced by `smooth_regularizer.m'.
%                   If Jh=[] or Jw=[], the regularization is ignored.
%            .Ww0- Initial column weight [w x r] (default [])
%                   Initilization by SVM is produced by `initw.m'.
%                   If Ww0=[] , it is randomly initialized.
%            .svmsolver - Linear svm solver, formed by "[alpha rho cost] = func(y,col_vectors,C)" [function_handle] (default @svmbysmo)
%		                   The bilinear optimization requires inner svm optimization (see sec 2.2 in the paper).
%						   You can give your own svm solver in this parameter field.
%            .verbose   - Verbose level, 0: sicent, 1: verbose (default 0) 
%            .rescale   - Rescaling or not [true|false] (default true)
%
% Output:
%  model  - Classifier model
%           .W  - Matrix weights [h x w],  W = Wh*Ww'
%           .rho- Bias [scalar], y = trace(W'*X) -rho
%           .Wh - Vertical (row) weights [h x r]
%           .Ww - Horizontal (column) weights [w x r]
%  status - Optimization status
%
%
% Reference:
%   T. Kobayashi and N. Otsu,
%   "Efficient Optimization For Low-Rank Integrated Bilinear Classifiers,"
%   Proc. European Conference on Computer Vision (ECCV), pp. 474-487, 2012.
%

  %- Parameters -%
  if ~exist('parms', 'var') || isempty(parms), parms = struct; end
  parms = parseparam(parms, 'C',1, 'Jh',[], 'Jw',[], 'Ww0',[], 'svmsolver',@svmbysmo, 'verbose',0, 'rescale',true);
  
  [parms.h, parms.w, parms.n] = size(X);
  
  %- Parameters in projected gradient descent -%
  parms = defaultparams(parms);

  %- Preprocess -%
  Uh = []; Uw = [];
  if ~isempty(parms.Jh)
  	[Uh Sh] = svd(eye(parms.h)+parms.Jh);
  	Uh = bsxfun(@times, Uh, 1./sqrt(diag(Sh)'));
	X = reshape(Uh' * reshape(X, parms.h, []), parms.h, parms.w, parms.n);
  end
  Xw = reshape(permute(X,[2 1 3]), parms.w, parms.h*parms.n); %[w x (h*n)]
  if ~isempty(parms.Jw)
  	[Uw Sw] = svd(eye(parms.w)+parms.Jw);
  	Uw = bsxfun(@times, Uw, 1./sqrt(diag(Sw)'));
	Xw = Uw' * Xw;
  end

  %- Init -%
  if isempty(parms.Ww0)
	  Ww0 = rand(parms.w, min(parms.h,parms.w));
	  Ww0 = Ww0/sum(Ww0(:));
  else
	  Ww0 = parms.Ww0;
  end

  %- Optimization by projected gradient -%
  [curpt,status] = PGDoptimize(Ww0, Xw, y, parms);

  %- Postprocess -%
  if ~isempty(Uh)
  	curpt.Wh = Uh*curpt.Wh;
  end
  if ~isempty(Uw)
  	curpt.Ww = Uw*curpt.Ww;
  end

  model.W = curpt.Wh*curpt.Ww';
  [oWh oS oWw] = svd(model.W,'econ'); oS = diag(oS);
  nzidx = (oS >= oS(1)*1e-7);
  if parms.verbose
	  fprintf('obj %f,  rank %d\n', curpt.cost, nnz(nzidx)); 
  end
  model.Wh  = bsxfun(@times, oWh(:,nzidx), sqrt(oS(nzidx)'));
  model.Ww  = bsxfun(@times, oWw(:,nzidx), sqrt(oS(nzidx)'));
  model.rho = curpt.b;
end

%= Optimization =%
function [curpt,diagnostic] = PGDoptimize(Ww0, Xw, y, parms)
  %- Init -%
  curpt = LIBsolvesubprob(Ww0, Xw, y, parms); %SVM
  [curpt.grad curpt.pgm2] = gradsubprob(curpt, Xw, parms); %gradient of Ww

  iter  = 0;
  step  = 1;
  evals = 1; % Used during initialisation
  BOOLstep0 = false; %check if the step size is too small
  if parms.verbose, p = plot(curpt.cost,'bo-'); set(gca,'yscale','log'); title(sprintf('cost: %f',curpt.cost)); end
  stoptol= 1e-9*curpt.pgm2;
  pgdtime = tic;
  while (curpt.pgm2 > stoptol && iter < parms.MAXITER && evals < parms.MAXEVAL && ~BOOLstep0)
	if step > parms.SQRBETAUP
		step = parms.BETADN; 
	end; 
    [newpt,step,subiter,BOOLstep0,evals] = steparmijo(curpt, Xw, y, parms, step, evals); %decrease step size
    if (parms.TYPEstep == 1) && (subiter == 1)
      [curpt,step,evals] = stepvararmijo(curpt, newpt, Xw, y, parms, step, evals); %increase step size
    else 
	  curpt = newpt; 
    end;
    iter = iter+1;
	if parms.verbose, set(p,'YData',[get(p,'YData') curpt.cost]); title(sprintf('cost: %f',curpt.cost)); drawnow; end
  end
  elpst = toc(pgdtime);
  BOOLconv    = (curpt.pgm2 <= stoptol);
  BOOLtimeout = ((iter >= parms.MAXITER) || (evals >= parms.MAXEVAL));
  diagnostic  = buildiagnostic(step,iter,evals,elpst,BOOLconv,BOOLtimeout,BOOLstep0);
end

%- line search for d by decreasing step size -%
function [newpt,step,iter,BOOLstep0,evals] = steparmijo(curpt, Xw, y, parms, step, evals)
  iter = 0;
  newpt.cost = +Inf;
  rhs = parms.SIGMATOL*curpt.pgm2;
  while ((newpt.cost > curpt.cost-step*rhs) && (iter < parms.MAXSUBEVAL) && (evals < parms.MAXEVAL))
    Ww    = PSDsvd(curpt.WwWw - step*curpt.grad);
    newpt = LIBsolvesubprob(Ww, Xw, y, parms);
    iter  = iter+1;
	evals = evals+1;
    step  = step*parms.BETADN; %decrease step size
  end
  step      = step/parms.BETADN; % Due to one extra iteration
  BOOLstep0 = (iter >= parms.MAXSUBEVAL);
  if newpt.cost < curpt.cost
    [newpt.grad newpt.pgm2] = gradsubprob(newpt, Xw, parms);
  else 
	newpt = curpt; 
  end
end

%- further line search for d by increasing step size -%
function [valpt,step,evals] = stepvararmijo(curpt, valpt, Xw, y, parms, step, evals)
  iter = 0;
  BOOLloop = true;
  rhs = parms.SIGMATOL*curpt.pgm2;
  while ((BOOLloop) && (evals < parms.MAXEVAL))
    iter     = iter+1;
	evals    = evals+1;
    newstep  = step*parms.BETAUP; %increase step size
    newWw    = PSDsvd(curpt.WwWw - newstep*curpt.grad);
    newpt    = LIBsolvesubprob(newWw, Xw, y, parms);
    BOOLloop = (newpt.cost <= curpt.cost-newstep*rhs);
	if BOOLloop
		valpt = newpt; 
		step  = newstep; 
	end;
  end;
  [valpt.grad valpt.pgm2] = gradsubprob(valpt, Xw, parms);
end

%- SVM optimization -%
function curpt = LIBsolvesubprob(Ww, Xw, y, parms)
  if all(abs(Ww)<1e-5)  %tiny weights
    curpt = struct('cost',Inf, 'alphay',[], 'b',Inf, 'Ww',Ww, 'Wh',[], 'WwWw',[]);
  else
	R  = size(Ww, 2);
	X_ = Ww'*Xw; %[R x h*n]
	X_ = reshape(X_, R*parms.h, parms.n); %[R*h x n]
	[alpha rho cost]= parms.svmsolver(y, X_, parms.C);
	curpt.cost   = cost + 0.5*sum(Ww(:).^2);
    curpt.alphay = alpha.*y;
    curpt.b      = rho;
    curpt.Ww     = Ww;
	Wh = reshape(X_*(curpt.alphay), R, parms.h)';
	curpt.Wh     = Wh; %[h x R]

    if all(alpha < eps)
       warning('PGDlearn2class:ZeroD','Numerical error [reached the origin].'); 
    end;
    
	if parms.rescale
		hmag2 = sum(Wh(:).^2);
		wmag2 = sum(Ww(:).^2);
		rescale = sqrt(hmag2./wmag2);
		
		curpt.cost   = curpt.cost - 0.5*(wmag2 + hmag2) + sqrt(wmag2*hmag2);
		curpt.alphay = curpt.alphay/rescale;
		curpt.Ww = sqrt(rescale)*Ww;
		curpt.Wh = Wh/sqrt(rescale); %[h x R]
	end;
	curpt.WwWw = curpt.Ww*curpt.Ww';
  end;
end

%= Calculate the gradient =%
function [grad pgm2] = gradsubprob(curpt, Xw, parms)
  [h w n] = deal(parms.h, parms.w, parms.n);
  
  idx  = abs(curpt.alphay) > eps;
  Xh   = reshape(Xw, w*h, n); %[w*h x n]
  X_   = reshape(Xh(:,idx) * curpt.alphay(idx), w, h); %[w x h]
  grad = 0.5*(eye(w) - X_*X_');
  pgm2 = sum(sum(grad.^2));
end

%= Diagnosis for optimization status =%
function diagnostic = buildiagnostic(step,iter,evals,deltat,BOOLconverged,BOOLtimeout,BOOLstep0)
  diagnostic.step  = step;
  diagnostic.iter  = iter;
  diagnostic.evals = evals;
  diagnostic.time  = deltat;
  diagnostic.BOOLstep0    = BOOLstep0;
  diagnostic.BOOLtimeout  = BOOLtimeout;
  diagnostic.BOOLconverged= BOOLconverged;
end

%= Projection onto semi-definite cone =%
function Ww = PSDsvd(W)
  [Ww S] = eig(W); Ww = real(Ww); S = real(diag(S));
  nzidx  = (S >= max(S)*1e-7) & (S > 0);
  Ww     = bsxfun(@times, Ww(:,nzidx), sqrt(S(nzidx)'));
  if isempty(Ww)
	  Ww = zeros(size(W,1),1);
  end
end

%= Default parameters for projected gradient =%
function parms = defaultparams(parms)
  % Multiple Kernel Learning parameters
  parms.BOOLMAP=false;         % Maximum A Posteirori estimate (not tested)
  
  % Gradient descent parameters
  parms.TYPEstep=1;           % 0 = Armijo, 1 = Variant Armijo

  parms.MAXITER=80;           % Maximum number of gradient descent iterations
%  parms.MAXITER=40;           % Maximum number of gradient descent iterations
  parms.MAXEVAL=200;          % Maximum number of SVM evaluations
  parms.MAXSUBEVAL=20;        % Maximum number of SVM evaluations
                              % modified according to manik's email

  parms.SIGMATOL=0.001;         % Needed by Armijo, variant Armijo for line search
%  parms.SIGMATOL=0.3;         % Needed by Armijo, variant Armijo for line search
  parms.BETAUP=2.1;           % Needed by Armijo, variant Armijo for line search
  parms.BETADN=0.3;           % Needed by variant Armijo for line search
  parms.SQRBETAUP=parms.BETAUP*parms.BETAUP;
end
