function [model status] = hMKLtrain(y, K, parms)
%
% Heterogeneous multiple kernel learning
% [model status] = hMKLtrain(y, K, params)
%
% Input:
%  y - Label vector {-1,+1}     [n x 1]
%  K - Multiple kernel matrices [n x n x #kernel]
%  params - Parameters for `bilinsvmtrain.m' [struct]
%            .C  - Balancing parameter   [scalar] (default 1)
%            .Jw - Constraint matrix for kernel weights [w x w] (default [])
%                   If Jw=[], the regularization is ignored.
%            .svmsolver - Linear svm solver, formed by func(y,col_vectors,C) [function_handle] (default @svmbysmo)
%		                  The bilinear optimization requires inner svm optimization (see sec 2.2 in the paper).
%			 			 You can give your own svm-solver in this parameter field.
%            .verbose   - Verbose level, 0: sicent, 1: verbose (default 0) 
%
% Output:
%  model  - Classifier model
%            .W   - Matrix weights [n x #kernel],  W = Wh*Ww'
%            .rho - Bias [scalar]
%            .Wh  - Vertical (row) weights [n x r]
%            .Ww  - Horizontal (column) weights [#kernel x r]
%            .hinvK - K^(-1/2) [n x n x #kernel]
%  status - Optimization status
%
%
% Reference:
%   T. Kobayashi and N. Otsu,
%   "Efficient Optimization For Low-Rank Integrated Bilinear Classifiers,"
%   Proc. European Conference on Computer Vision (ECCV), pp. 474-487, 2012.
%

[n, m, nK] = size(K);
if n ~= m
	error('hMKLtrain: wrong size of kernel matrices.');
end
X = zeros(n, nK, n);
hinvK   = zeros(n, n, nK);
for i = 1:nK
	[V S] = svd(K(:,:,i)); S = diag(S);
	nzidx = S >= S(1)*1e-7;
	V = bsxfun(@times, V(:,nzidx), 1./sqrt(sqrt(S(nzidx)')));
	hinvK(:,:,i) = V*V';
	X(:,i,:) = hinvK(:,:,i)*K(:,:,i);
end

%- Init by svm -%
Ww0= initw(y, X, struct('svmsolver',@svmbysmo));
parms.Ww0 = Ww0;

[model status] = bilinsvmtrain(y, X, parms);
model.hinvK = hinvK;
