function J = smooth_regularizer(dims, ord)
%
% Regularizer for smoothness in quadratic form x'Jx
%  J = smooth_regularizer(dims, ord)
%
% Input:
%  dims - Dimensions; for 1-D signal, [n], for 2-D position, [h x w]
%  ord  - Order for smoothing ['1st'|'2nd']
%         '1st' : 1st order derivatives, sum_i (x_i - x_{i+1})^2
%         '2nd' : 2nd order derivatives, sum_i (x_{i-1} - 2*x_i + x_{i+1})^2
%
% Output:
%  J - Regularization matrix [n x n] or [h*w x h*w]
%
%

if numel(dims) == 1 
	%= 1-dimensional =%
	switch(ord)
	case '2nd' %2-nd derivatives
		diag1 = repmat(1.5,dims,1); diag1([1,2,end-1,end]) = [1.25, 2.25, 2.25, 1.25];
		diag2 = -ones(dims-1,1);    diag2([1,end]) = -[1.5, 1.5];
		diag3 = repmat(0.25,dims-2,1);
		J = diag(diag1) + diag(diag2,1) + diag(diag2,-1) + diag(diag3,2) + diag(diag3,-2);
	case '1st' %1-st derivatives
		J = diag(repmat(2,dims,1)) - diag(ones(dims-1,1),1) - diag(ones(dims-1,1),-1);
		J(1,1) = 1;
		J(end,end) = 1;
	otherwise
		error('wrong order.');
	end
elseif numel(dims) == 2
	%= 2-dimensional =%
	switch(ord)
	case '2nd'
		Map = zeros(dims(1),dims(2),prod(dims));
		tmpl= [1 2 1;2 0 2;1 2 1];
		count = 1;
		for i = 1:dims(1)
			for j = 1:dims(2)
				iind = max(i-1,1):min(i+1,dims(1));
				jind = max(j-1,1):min(j+1,dims(2));
				tmpl_= tmpl(iind-i+2,jind-j+2);
				Map(iind,jind,count) = tmpl_/sum(tmpl_(:));
				Map(i,j,count) = -1;
				count = count + 1;
			end
		end
		J = reshape(Map,[],count-1)*reshape(Map,[],count-1)';
	case '1st'
		Mapx = zeros(dims(1),dims(2),(dims(1)-1)*(dims(2)-1));
		Mapy = zeros(dims(1),dims(2),(dims(1)-1)*(dims(2)-1));
		count = 1;
		for i = 1:dims(1)-1
			for j = 1:dims(2)-1
				Mapx(i:i+1,j:j+1,count) = [1 -1;1 -1];
				Mapy(i:i+1,j:j+1,count) = [1 1;-1 -1];
				count = count + 1;
			end
		end
		J = reshape(Mapx,[],count-1)*reshape(Mapx,[],count-1)' + reshape(Mapy,[],count-1)*reshape(Mapy,[],count-1)';
	otherwise
		error('wrong order.');
	end
else
	error('wrong dims. (be <= 2)');
end
